/*
 * Receive batched mail - Rewritten for OS/2 by Dmitry Zavalishin
 * Original version was made by avg
 *
 * Patched by Thomas J. Dire, 6.11.98.
 * (FAQ version of spawn call)
 *
 *    $Log: rbmail.c $
 * Revision 1.3  1995/03/10  03:37:48  dz
 * cleanup
 *
 * Revision 1.2  1995/03/06  13:40:49  dz
 * - don't die if wrong command line given, just process batch
 * - don't flood system with a lots of rmails
 *
 *
 */

#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include <time.h>
#include <process.h>
#include <ctype.h>


#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_DOSQUEUES
#define INCL_DOSFILEMGR
#include <OS2.H>

const max_header = 4096;

int	mail( char * const args[] );
void	dump_off( const char *err_msg, const char *hdr, const char* data, int len );
void	corrupt_hdr( const char *details, const char *hdr );
int	send_it( const char* data, int len, char * const args[] );



typedef const char *	cchar_ptr;
typedef cchar_ptr &	cchar_ptr_ref;


char *
get_one_arg( cchar_ptr_ref line, char *reason )
	{
	char *out;
	const char *begin;

// printf( "Get one Arg got '%s'\n", line );

	begin = line;
	while( *line && !isspace(*line) ) line ++;

	int len = line - begin;

	out = (char *) malloc( len + 1 );

	if( out == NULL )
		{
		strcpy( reason, "Out of memory" );
		return NULL;
		}

	strncpy( out, begin, len );
	out[len] = '\0';

	return out;
	}


int
load_args( char *rargs[], const char *line, char *fail_reason )
	{
	int arg = 1;

	rargs[0] = "rmail"; // this is Tom Dire's change, 6.XI.1998

// printf( "LoadArgs got '%s'\n", line );

	while(1)
		{

		while( isspace( *line ) ) line++;

		if( *line == '\0' )
			break;

		rargs[arg] = get_one_arg( line, fail_reason );
		if( rargs[arg] == NULL )
			return -1;

		arg++;

		if( arg >= 98 )
			{
			strcpy( fail_reason, "Too many args" );
			return -1;
			}
		}

	rargs[arg] = NULL;

/*
for( int i = 0; i < 99; i++ )
  if( rargs[i] == NULL ) break;
  else printf("rarg[%d] = '%s', ", i, rargs[i] );
printf("\n");
*/
	return 0;
	}



void
do_one_piece( const char *headerline )
	{
	char diagheader[max_header];
	char *rargs[100];
	const char *p;
	unsigned rcsum, csum;
	long     len;

	//	Parse header line
	rargs[0] = NULL;
	strcpy(diagheader, headerline);

	/* Length */
	p = headerline;
	len = 0;
	for( int i = 0 ; i < 7 ; i++ )
		{
		unsigned char c = *p++;
		if( '0' > c || c > '9' )
			corrupt_hdr( "bad length", diagheader );
		len = (len*10) + (c-'0');
		}

	if( *p++ != ' ' )
		corrupt_hdr( "no space after length", diagheader );

	/* Checksum */
	rcsum = 0;
	for( i = 0 ; i < 4 ; i++ )
		{
		unsigned char c = *p++;
		rcsum <<= 4;
		if( '0' <= c && c <= '9' )
			rcsum |= c-'0';
		else if( 'A' <= c && c <= 'F' )
			rcsum |= c-'A'+0xa;
		else if( 'a' <= c && c <= 'f' )
			rcsum |= c-'a'+0xa;
		else
			corrupt_hdr( "bad checksum field", diagheader );
		}

	if( *p != ' ' )
		corrupt_hdr( "no space after checksum", diagheader );

	/* Destination addresses */
	if( ! *p )
		corrupt_hdr( "no dest address", diagheader );


	char reason[200];
	if( load_args( rargs, p, reason ) != 0 )
		corrupt_hdr( reason, diagheader );

/*
for( int j = 0; j < 99; j++ )
  if( rargs[j] == NULL ) break;
  else printf("rarg[%d] = '%s', ", j, rargs[j] );
printf("\n");
*/

	// Read all the stuff in

	char *Letter = (char *) malloc( len+1 );

	if( Letter == NULL )
		corrupt_hdr("Out of memory", diagheader );

	int real_len = 0;
	if( (real_len = fread( Letter, 1, len, stdin )) != len )
		dump_off( "Can't read data", diagheader, Letter, real_len );

	Letter[len] = '\0';

	// Calc. checksum

	csum = 0;
	for( char *cp = Letter; *cp; cp++ )
		{
		if( csum & 01 )
			csum = (csum>>1) + 0x8000;
		else
			csum >>= 1;
		csum += *cp;
		csum &= 0xFFFF;
		}

//	if( rcsum != csum )
//		dump_off( "Checksum error", diagheader, Letter, real_len );

	if( send_it( Letter, real_len, rargs ) != 0 )
		dump_off( "Can't send", diagheader, Letter, real_len );

	for( char **ap = &rargs[1]; *ap; ap++ )
		free( *ap );

	free( Letter );
	}


int
main( int ac, char *av[] )
	{
	char headerline[max_header];

	if( ac != 1 )
		{
		fprintf(stderr, "Wrong arg: %s\n\n", av[1] );
		fprintf(stderr, "Usage: rbmail < batchfile\n");

// never give up! ᫨  p㪠 ⨫ - p㪠  p訡,
//  ᫨ uuxqt p    稪 譨 -     
//   㤥 p   p檨 .
//		exit(64);       /* EX_USAGE */
		}

	while( fgets(headerline, (sizeof headerline)-1, stdin) != NULL )
		{
		headerline[(sizeof headerline)-1] = '\0';

		char *p = headerline;
		while( *p != '\n' && *p ) p++;
		if( ! *p )
			corrupt_hdr( "missing newline", headerline );
		*p = '\0';

		do_one_piece( headerline );
		}

	return 0;
	}



//
// Spawn rmail, return descriptor of our end of it's pipe
//

int
mail( char * const args[] )
	{
	unsigned long rd, wr;

/*
for( int i = 0; i < 99; i++ )
  if( args[i] == NULL ) break;
  else printf("arg[%d] = '%s', ", i, args[i] );
printf("\n");
*/

	int stdin_save = dup( 0 );

	if ( DosCreatePipe( &rd, &wr, 4096 ) != 0 )
		return -1;

	if( 0 != dup2( rd, 0 ) )
		return -1;

	close( rd );
        DosSetFHState( wr, OPEN_FLAGS_NOINHERIT );

	if( spawnvp( P_NOWAITO, "rmail", args ) < 0 )
		return -1;

	if( 0 != dup2( stdin_save, 0 ) )
		return -1;

	close( stdin_save );

	return wr;
	}


void
dump_off( const char *err_msg, const char *hdr, const char* data, int len )
	{
	char msg[1000];

	char *dir;

	if( NULL != (dir = getenv("RBMAIL")) )
		if( NULL != (dir = getenv("TMP")) )
			if( NULL != (dir = getenv("TEMP")) )
				dir = "\\";

	chdir( dir );
	mkdir( "Errors" );
	chdir( "Errors" );

	char *tname = "ErXXXXXX";
	mktemp(tname);

	FILE *ef = fopen( tname, "wb" );

	if( ef == NULL )
		{
		fprintf( stderr, "Can't save broken batch!\n" );
		exit(44);
		}

	fprintf( ef, "Error: %s\n%s\n", err_msg, hdr );
	fwrite( data, 1, len, ef );

	const bs = 16*1024;
	char buf[bs];
	int i;

	while( (i = fread( buf, 1, bs, stdin )) > 0 )
		fwrite( buf, 1, i, ef );

	int ee = ferror( ef );		// Error writing ef?

	fclose( ef );

	char *margs[3];

	margs[0] = "rmail";
	margs[1] = "postmaster";
	margs[2] = NULL;

	int pipe_d = mail( margs );	// Spawn rmail, return pipe to it
	FILE *mf = fdopen( pipe_d, "wb" );

	if( pipe_d < 0 || mf == NULL )
		{
		fprintf( stderr, "Can't mail to postmaster!\n" );
		exit(44);
		}

	long tt = time(NULL);

	fprintf( mf, "From uucp %s", ctime(&tt));
	fprintf( mf, "From: uucp (UUCP Batch Mail Service)\n");
	fprintf( mf, "To: postmaster\n");
	fprintf( mf, "Subject: batch receiption failed\n\n");
	fprintf( mf, "ERROR: %s.\nBatch saved in %s\n\n", err_msg, tname );

	if( ee )
		fprintf( mf, "There was error writing %s\n\n", tname );

	fclose( mf );
	DosClose( pipe_d );

	exit( 33 );
	}


void
corrupt_hdr( const char *details, const char *hdr )
	{
	char msg[1000];
	sprintf( msg, "Header line corrupted (%s)", details );

	// dump ther rest of file, send error message

	dump_off( msg, hdr, NULL, 0 );

	exit( 33 );
	}



void
be_idle_a_bit( void )
	{
	static int	turn = 0;

//  ⠪:
//    ⮫쪮  ⨫ 뭮 p .
//   ⠪ ⠢  ᥡ idle priority, ⮡ pp ᥡ,
//   ᫨ p㧪 設 ,   ᪠   ⥩.

	fprintf( stderr, "rbmail: Sleeping\n");
	DosSleep( 1000 );
	fprintf( stderr, "rbmail: Idle\n");
	DosSetPriority (PRTYS_THREAD, PRTYC_IDLETIME, 15L, 0L );
	DosSleep( 500 );
	DosSleep( 500 );
	DosSleep( 500 );
	DosSleep( 500 );
	DosSetPriority (PRTYS_THREAD, PRTYC_REGULAR, 0L, 0L );
	fprintf( stderr, "rbmail: Regular priority\n");



	if( ++turn > 9 )
		{
		PID		ch_pid;
		RESULTCODES	ch_rc;
//
//  㦥 p ⮪ pᮢ -  ⮪!
//

		fprintf( stderr, "rbmail: waiting for some child to end\n");
		DosWaitChild( DCWA_PROCESSTREE, DCWW_WAIT, &ch_rc, &ch_pid, 0 );
		fprintf( stderr, "rbmail: some child finished\n");

		turn--;
		}
	}



int
send_it( const char* data, int len, char * const args[] )
	{
	int pipe_d = mail( args );	// Spawn rmail, return pipe to it
	unsigned long actual;

	if( pipe_d < 0 )
		{
		fprintf( stderr, "Can't mail!\n" );
		return -1;
		}

	DosWrite( pipe_d, (void *)data, len, &actual );

	DosClose( pipe_d );

	be_idle_a_bit();

	return actual == len ? 0 : -1;
	}


